@@ -55,6 +55,9 @@
 #include <openssl/rsa.h>
 #include <openssl/x509v3.h>
 #include <openssl/err.h>
+#include <openssl/evp.h>
+#include <openssl/crypto.h>
+#include <zlib.h>
 
 #include "cmdline.h"
 #include "../common/util.h"
@@ -119,8 +122,25 @@ static bool sign_data(ykpiv_state *state, const unsigned char *in, size_t len, u
 
   unsigned char signinput[1024] = {0};
   if(YKPIV_IS_RSA(algorithm)) {
-    size_t padlen = algorithm == YKPIV_ALGO_RSA1024 ? 128 : 256;
-    if(RSA_padding_add_PKCS1_type_1(signinput, padlen, in, len) == 0) {
+    size_t padlen = 0;
+    switch (algorithm) {
+      case YKPIV_ALGO_RSA1024:
+        padlen = 128;
+        break;
+      case YKPIV_ALGO_RSA2048:
+        padlen = 256;
+        break;
+      case YKPIV_ALGO_RSA3072:
+        padlen = 384;
+        break;
+      case YKPIV_ALGO_RSA4096:
+        padlen = 512;
+        break;
+      default:
+        fprintf(stderr, "Unknown RSA algorithm.\n");
+        return false;
+    }
+    if (RSA_padding_add_PKCS1_type_1(signinput, padlen, in, len) == 0) {
       fprintf(stderr, "Failed adding padding.\n");
       return false;
     }
@@ -217,7 +237,7 @@ static EVP_PKEY* wrap_public_key(ykpiv_state *state, int algorithm, EVP_PKEY *pu
     }
     EVP_PKEY_assign_RSA(pkey, sk);
   }
-  else {
+  else if(YKPIV_IS_EC(algorithm)){
     const EC_KEY *ec = EVP_PKEY_get0_EC_KEY(public_key);
     EC_KEY_METHOD *meth = EC_KEY_METHOD_new(EC_KEY_get_method(ec));
     EC_KEY_METHOD_set_init(meth, NULL, yk_ec_meth_finish, NULL, NULL, NULL, NULL);
@@ -239,6 +259,19 @@ static EVP_PKEY* wrap_public_key(ykpiv_state *state, int algorithm, EVP_PKEY *pu
 }
 #endif
 
+static bool move_key(ykpiv_state *state, int from_slot, int to_slot) {
+  bool ret = false;
+  ykpiv_rc res;
+
+  res = ykpiv_move_key(state, (uint8_t) (from_slot & 0xFF), (uint8_t) (to_slot & 0xFF));
+  if (res != YKPIV_OK) {
+    fprintf(stderr, "Failed to move key.\n");
+  } else {
+    ret = true;
+  }
+  return ret;
+}
+
 static bool generate_key(ykpiv_state *state, enum enum_slot slot,
     enum enum_algorithm algorithm, const char *output_file_name,
     enum enum_key_format key_format, enum enum_pin_policy pin_policy,
@@ -287,62 +320,85 @@ static bool generate_key(ykpiv_state *state, enum enum_slot slot,
     goto generate_out;
   }
 
-  if(key_format == key_format_arg_PEM) {
+  if (key_format == key_format_arg_PEM) {
     public_key = EVP_PKEY_new();
-    if(algorithm == algorithm_arg_RSA1024 || algorithm == algorithm_arg_RSA2048) {
-      BIGNUM *bignum_n = NULL;
-      BIGNUM *bignum_e = NULL;
-      rsa = RSA_new();
-      bignum_n = BN_bin2bn(mod, mod_len, NULL);
-      if (bignum_n == NULL) {
-        fprintf(stderr, "Failed to parse public key modulus.\n");
-        goto generate_out;
-      }
-      bignum_e = BN_bin2bn(exp, exp_len, NULL);
-      if(bignum_e == NULL) {
-        fprintf(stderr, "Failed to parse public key exponent.\n");
-        goto generate_out;
-      }
+    switch (algorithm) {
+      case algorithm_arg_RSA1024:
+      case algorithm_arg_RSA2048:
+      case algorithm_arg_RSA3072:
+      case algorithm_arg_RSA4096: {
+        BIGNUM *bignum_n = NULL;
+        BIGNUM *bignum_e = NULL;
+        rsa = RSA_new();
+        bignum_n = BN_bin2bn(mod, mod_len, NULL);
+        if (bignum_n == NULL) {
+          fprintf(stderr, "Failed to parse public key modulus.\n");
+          goto generate_out;
+        }
+        bignum_e = BN_bin2bn(exp, exp_len, NULL);
+        if (bignum_e == NULL) {
+          fprintf(stderr, "Failed to parse public key exponent.\n");
+          goto generate_out;
+        }
 
-      if(RSA_set0_key(rsa, bignum_n, bignum_e, NULL) != 1) {
-        fprintf(stderr, "Failed to set RSA key\n");
-        goto generate_out;
+        if (RSA_set0_key(rsa, bignum_n, bignum_e, NULL) != 1) {
+          fprintf(stderr, "Failed to set RSA key\n");
+          goto generate_out;
+        }
+        if (EVP_PKEY_set1_RSA(public_key, rsa) != 1) {
+          fprintf(stderr, "Failed to set RSA public key\n");
+          goto generate_out;
+        }
       }
-      if(EVP_PKEY_set1_RSA(public_key, rsa) != 1) {
-        fprintf(stderr, "Failed to set RSA public key\n");
-        goto generate_out;
-      }
-    } else if(algorithm == algorithm_arg_ECCP256 || algorithm == algorithm_arg_ECCP384) {
-      int nid;
+        break;
+      case algorithm_arg_ECCP256:
+      case algorithm_arg_ECCP384: {
+        int nid;
 
-      if(algorithm == algorithm_arg_ECCP256) {
-        nid = NID_X9_62_prime256v1;
-      } else {
-        nid = NID_secp384r1;
-      }
-      eckey = EC_KEY_new();
-      group = EC_GROUP_new_by_curve_name(nid);
-      EC_GROUP_set_asn1_flag(group, OPENSSL_EC_NAMED_CURVE);
-      if(EC_KEY_set_group(eckey, group) != 1) {
-        fprintf(stderr, "Failed to set EC group.\n");
-        goto generate_out;
-      }
-      ecpoint = EC_POINT_new(group);
+        if (algorithm == algorithm_arg_ECCP256) {
+          nid = NID_X9_62_prime256v1;
+        } else {
+          nid = NID_secp384r1;
+        }
+        eckey = EC_KEY_new();
+        group = EC_GROUP_new_by_curve_name(nid);
+        EC_GROUP_set_asn1_flag(group, OPENSSL_EC_NAMED_CURVE);
+        if (EC_KEY_set_group(eckey, group) != 1) {
+          fprintf(stderr, "Failed to set EC group.\n");
+          goto generate_out;
+        }
+        ecpoint = EC_POINT_new(group);
 
-      if(!EC_POINT_oct2point(group, ecpoint, point, point_len, NULL)) {
-        fprintf(stderr, "Failed to load public point.\n");
-        goto generate_out;
+        if (!EC_POINT_oct2point(group, ecpoint, point, point_len, NULL)) {
+          fprintf(stderr, "Failed to load public point.\n");
+          goto generate_out;
+        }
+        if (!EC_KEY_set_public_key(eckey, ecpoint)) {
+          fprintf(stderr, "Failed to set the public key.\n");
+          goto generate_out;
+        }
+        if (EVP_PKEY_set1_EC_KEY(public_key, eckey) != 1) {
+          fprintf(stderr, "Failed to set EC public key.\n");
+          goto generate_out;
+        }
       }
-      if(!EC_KEY_set_public_key(eckey, ecpoint)) {
-        fprintf(stderr, "Failed to set the public key.\n");
-        goto generate_out;
-      }
-      if(EVP_PKEY_set1_EC_KEY(public_key, eckey) != 1) {
-        fprintf(stderr, "Failed to set EC public key.\n");
-        goto generate_out;
-      }
-    } else {
-      fprintf(stderr, "Wrong algorithm.\n");
+        break;
+#if (OPENSSL_VERSION_NUMBER >= 0x10100000L)
+      case algorithm_arg_ED25519:
+        public_key = EVP_PKEY_new_raw_public_key(EVP_PKEY_ED25519, NULL, point, point_len);
+        break;
+      case algorithm_arg_X25519:
+        public_key = EVP_PKEY_new_raw_public_key(EVP_PKEY_X25519, NULL, point, point_len);
+        break;
+#else
+      case algorithm_arg_ED25519:
+      case algorithm_arg_X25519:
+        fprintf(stderr, "Key was generated successfully but a public key cannot be parsed due to too old OpenSSL version. "
+                        "Upgrade OpenSSL to at least 1.1 or use attestation command to get a signed certificate instead.\n");
+        return true;
+#endif
+      default:
+        fprintf(stderr, "Wrong algorithm.\n");
     }
     if(PEM_write_PUBKEY(output_file, public_key) == 1) {
       ret = true;
@@ -460,16 +516,30 @@ static bool import_key(ykpiv_state *state, enum enum_key_format key_format,
     if(YKPIV_IS_RSA(algorithm)) {
       RSA *rsa_private_key = EVP_PKEY_get1_RSA(private_key);
       unsigned char e[4] = {0};
-      unsigned char p[128] = {0};
-      unsigned char q[128] = {0};
-      unsigned char dmp1[128] = {0};
-      unsigned char dmq1[128] = {0};
-      unsigned char iqmp[128] = {0};
+      unsigned char p[256] = {0};
+      unsigned char q[256] = {0};
+      unsigned char dmp1[256] = {0};
+      unsigned char dmq1[256] = {0};
+      unsigned char iqmp[256] = {0};
       const BIGNUM *bn_e, *bn_p, *bn_q, *bn_dmp1, *bn_dmq1, *bn_iqmp;
 
-      int element_len = 128;
-      if(algorithm == YKPIV_ALGO_RSA1024) {
-        element_len = 64;
+      int element_len = 0;
+      switch(algorithm) {
+        case YKPIV_ALGO_RSA1024:
+          element_len = 64;
+          break;
+        case YKPIV_ALGO_RSA2048:
+          element_len = 128;
+          break;
+        case YKPIV_ALGO_RSA3072:
+          element_len = 192;
+          break;
+        case YKPIV_ALGO_RSA4096:
+          element_len = 256;
+          break;
+        default:
+          fprintf(stderr, "Unsupported RSA algorithm\n");
+          goto import_out;
       }
 
       RSA_get0_key(rsa_private_key, NULL, &bn_e, NULL);
@@ -507,11 +577,11 @@ static bool import_key(ykpiv_state *state, enum enum_key_format key_format,
       }
 
       rc = ykpiv_import_private_key(state, key, algorithm,
-                                    p, (size_t)element_len,
-                                    q, (size_t)element_len,
-                                    dmp1, (size_t)element_len,
-                                    dmq1, (size_t)element_len,
-                                    iqmp, (size_t)element_len,
+                                    p, element_len,
+                                    q, element_len,
+                                    dmp1, element_len,
+                                    dmq1, element_len,
+                                    iqmp, element_len,
                                     NULL, 0,
                                     pp, tp);
     }
@@ -539,6 +609,26 @@ static bool import_key(ykpiv_state *state, enum enum_key_format key_format,
                                     s_ptr, element_len,
                                     pp, tp);
     }
+#if (OPENSSL_VERSION_NUMBER >= 0x10100000L)
+    else if(YKPIV_IS_25519(algorithm)) {
+      unsigned char s_ptr[48] = {0};
+      size_t element_len = sizeof(s_ptr);
+
+      if (EVP_PKEY_get_raw_private_key(private_key, s_ptr, &element_len) != 1) {
+        fprintf(stderr, "Failed to extract private key.\n");
+        goto import_out;
+      }
+
+      rc = ykpiv_import_private_key(state, key, algorithm,
+                                    NULL, 0,
+                                    NULL, 0,
+                                    NULL, 0,
+                                    NULL, 0,
+                                    NULL, 0,
+                                    s_ptr, element_len,
+                                    pp, tp);
+    }
+#endif
 
     ret = true;
     if(rc != YKPIV_OK) {
@@ -566,14 +656,14 @@ import_out:
   return ret;
 }
 
-static bool import_cert(ykpiv_state *state, enum enum_key_format cert_format,
+static bool import_cert(ykpiv_state *state, enum enum_key_format cert_format, int perform_compress,
     const char *input_file_name, enum enum_slot slot, char *password) {
   bool ret = false;
   FILE *input_file = NULL;
   X509 *cert = NULL;
   PKCS12 *p12 = NULL;
   EVP_PKEY *private_key = NULL;
-  int compress = 0;
+  int compress = YKPIV_CERTINFO_UNCOMPRESSED;
   int cert_len = -1;
 
   input_file = open_file(input_file_name, key_file_mode(cert_format, false));
@@ -619,7 +709,7 @@ static bool import_cert(ykpiv_state *state, enum enum_key_format cert_format,
       goto import_cert_out;
     }
     cert_len = st.st_size;
-    compress = 0x01;
+    compress = YKPIV_CERTINFO_GZIP;
   } else {
     /* TODO: more formats go here */
     fprintf(stderr, "Unknown key format.\n");
@@ -634,8 +724,8 @@ static bool import_cert(ykpiv_state *state, enum enum_key_format cert_format,
     unsigned char *certptr = certdata;
     ykpiv_rc res;
 
-    if(cert_len > YKPIV_OBJ_MAX_SIZE || cert_len < 0) {
-      fprintf(stderr, "Length of certificate is more than can fit.\n");
+    if((cert_len > YKPIV_OBJ_MAX_SIZE && !perform_compress) || cert_len < 0) {
+      fprintf(stderr, "Length of certificate is more than can fit. Consider using the 'compress' flag\n");
       goto import_cert_out;
     }
 
@@ -644,12 +734,44 @@ static bool import_cert(ykpiv_state *state, enum enum_key_format cert_format,
         fprintf(stderr, "Failed to read compressed certificate\n");
         goto import_cert_out;
       }
+    } else if(perform_compress) {
+      unsigned char uncompressed_certdata[YKPIV_OBJ_MAX_SIZE*10] = {0};
+      unsigned char *uncompressed_certptr = uncompressed_certdata;
+      if(i2d_X509(cert, &uncompressed_certptr) < 0) {
+        fprintf(stderr, "Failed to encode X509 certificate before compression\n");
+        goto import_cert_out;
+      }
+
+      z_stream zs;
+      zs.zalloc = Z_NULL;
+      zs.zfree = Z_NULL;
+      zs.opaque = Z_NULL;
+      zs.avail_in = (uInt)cert_len;
+      zs.next_in = (Bytef *)uncompressed_certdata;
+      zs.avail_out = (uInt) sizeof(certdata);
+      zs.next_out = (Bytef *)certdata;
+
+      if(deflateInit2(&zs, Z_DEFAULT_COMPRESSION, Z_DEFLATED, MAX_WBITS | 16, 8, Z_DEFAULT_STRATEGY) != Z_OK) {
+        fprintf(stderr, "Failed to compress certificate\n");
+        goto import_cert_out;
+      }
+      if(deflate(&zs, Z_FINISH) != Z_STREAM_END) {
+        fprintf(stderr, "Failed to compress certificate\n");
+        goto import_cert_out;
+      }
+      if(deflateEnd(&zs) != Z_OK) {
+        fprintf(stderr, "Failed to compress certificate\n");
+        goto import_cert_out;
+      }
+      cert_len = zs.total_out;
+      compress = YKPIV_CERTINFO_GZIP;
     } else {
       if(i2d_X509(cert, &certptr) < 0) {
         fprintf(stderr, "Failed to encode X509 certificate\n");
         goto import_cert_out;
       }
     }
+
     if ((res = ykpiv_util_write_cert(state, get_slot_hex(slot), certdata, (size_t)cert_len, compress)) != YKPIV_OK) {
       fprintf(stderr, "Failed commands with device: %s\n", ykpiv_strerror(res));
     } else {
@@ -744,12 +866,12 @@ static bool request_certificate(ykpiv_state *state, enum enum_key_format key_for
   FILE *input_file = NULL;
   FILE *output_file = NULL;
   EVP_PKEY *public_key = NULL;
-  const EVP_MD *md;
+  const EVP_MD *md = NULL;
   bool ret = false;
   unsigned char algorithm;
   int key = 0;
-  size_t oid_len;
-  const unsigned char *oid;
+  size_t oid_len = 0;
+  const unsigned char *oid = NULL;
 #if (OPENSSL_VERSION_NUMBER < 0x10100000L) || defined(LIBRESSL_VERSION_NUMBER)
   unsigned char digest[EVP_MAX_MD_SIZE + MAX_OID_LEN] = {0};
   unsigned int md_len;
@@ -835,10 +957,11 @@ static bool request_certificate(ykpiv_state *state, enum enum_key_format key_for
   if(algorithm == 0) {
     goto request_out;
   }
-
-  md = get_hash(hash, &oid, &oid_len);
-  if(md == NULL) {
-    goto request_out;
+  if (!YKPIV_IS_25519(algorithm)) {
+    md = get_hash(hash, &oid, &oid_len);
+    if (md == NULL) {
+      goto request_out;
+    }
   }
 
   if(!X509_REQ_set_pubkey(req, public_key)) {
@@ -904,17 +1027,61 @@ static bool request_certificate(ykpiv_state *state, enum enum_key_format key_for
     req->signature->flags = ASN1_STRING_FLAG_BITS_LEFT;
   }
 #else
-  /* With opaque structures we can not touch whatever we want, but we need
-   * to embed the sign_data function in the RSA/EC key structures  */
-  EVP_PKEY *sk = wrap_public_key(state, algorithm, public_key, key, oid, oid_len);
 
-  if(X509_REQ_sign(req, sk, md) == 0) {
-    fprintf(stderr, "Failed signing request.\n");
-    ERR_print_errors_fp(stderr);
+#if (OPENSSL_VERSION_NUMBER >= 0x10100000L)
+  if (algorithm == YKPIV_ALGO_ED25519) {
+
+    // Generate a dummy ED25519 to sign with OpenSSL
+    EVP_PKEY *ed_key = NULL;
+    EVP_PKEY_CTX *ed_ctx = EVP_PKEY_CTX_new_id(EVP_PKEY_ED25519, NULL);
+    EVP_PKEY_keygen_init(ed_ctx);
+    EVP_PKEY_keygen(ed_ctx, &ed_key);
+    EVP_PKEY_CTX_free(ed_ctx);
+
+    // Sign the request object using the dummy key
+    if (X509_REQ_sign(req, ed_key, md) == 0) {
+      fprintf(stderr, "Failed signing certificate.\n");
+      ERR_print_errors_fp(stderr);
+      EVP_PKEY_free(ed_key);
+      goto request_out;
+    }
+    EVP_PKEY_free(ed_key);
+
+    // Extract the request data without the signature
+    unsigned char *tbs_data = NULL;
+    int tbs_len = i2d_re_X509_REQ_tbs(req, &tbs_data);
+
+    // Sign the request data using the YubiKey
+    unsigned char yk_sig[64] = {0};
+    size_t yk_siglen = sizeof(yk_sig);
+    if (!sign_data(state, tbs_data, tbs_len, yk_sig, &yk_siglen, algorithm, key)) {
+      fprintf(stderr, "Failed signing tbs request portion.\n");
+      goto request_out;
+    }
+
+    // Replace the dummy signature with the signature from the yubikey
+    ASN1_BIT_STRING *psig;
+    const X509_ALGOR *palg;
+    X509_REQ_get0_signature(req, (const ASN1_BIT_STRING **) &psig, &palg);
+    ASN1_BIT_STRING_set(psig, yk_sig, yk_siglen);
+
+  } else {
+#endif
+    /* With opaque structures we can not touch whatever we want, but we need
+     * to embed the sign_data function in the RSA/EC key structures  */
+    EVP_PKEY *sk = wrap_public_key(state, algorithm, public_key, key, oid, oid_len);
+
+    if(X509_REQ_sign(req, sk, md) == 0) {
+      fprintf(stderr, "Failed signing request.\n");
+      ERR_print_errors_fp(stderr);
+      EVP_PKEY_free(sk);
+      goto request_out;
+    }
     EVP_PKEY_free(sk);
-    goto request_out;
+#if (OPENSSL_VERSION_NUMBER >= 0x10100000L)
   }
-  EVP_PKEY_free(sk);
+#endif
+
 #endif
 
   if(key_format == key_format_arg_PEM) {
@@ -923,6 +1090,12 @@ static bool request_certificate(ykpiv_state *state, enum enum_key_format key_for
     } else {
       fprintf(stderr, "Failed writing x509 information\n");
     }
+  } else if(key_format == key_format_arg_DER) {
+    if(i2d_X509_REQ_fp(output_file, req)) {
+      ret = true;
+    } else {
+      fprintf(stderr, "Failed writing DER information\n");
+    }
   } else {
     fprintf(stderr, "Only PEM support available for certificate requests.\n");
   }
@@ -1040,12 +1213,19 @@ static bool selfsign_certificate(ykpiv_state *state, enum enum_key_format key_fo
   if(algorithm == 0) {
     goto selfsign_out;
   }
+  if(algorithm == YKPIV_ALGO_X25519) {
+    fprintf(stderr, "Signing with X25519 keys is not supported.\n");
+    goto selfsign_out;
+  }
 
   size_t oid_len = 0;
   const unsigned char *oid = 0;
-  const EVP_MD *md = get_hash(hash, &oid, &oid_len);
-  if(md == NULL) {
-    goto selfsign_out;
+  const EVP_MD *md = NULL;
+  if (algorithm != YKPIV_ALGO_ED25519) {
+    md = get_hash(hash, &oid, &oid_len);
+    if (md == NULL) {
+      goto selfsign_out;
+    }
   }
   x509 = X509_new();
   if(!x509) {
@@ -1207,18 +1387,60 @@ static bool selfsign_certificate(ykpiv_state *state, enum enum_key_format key_fo
     x509->signature->flags = ASN1_STRING_FLAG_BITS_LEFT;
   }
 #else
-  /* With opaque structures we can not touch whatever we want, but we need
-   * to embed the sign_data function in the RSA/EC key structures  */
-  EVP_PKEY *sk = wrap_public_key(state, algorithm, public_key, key, oid, oid_len);
 
-  if(X509_sign(x509, sk, md) == 0) {
-    fprintf(stderr, "Failed signing certificate.\n");
-    ERR_print_errors_fp(stderr);
+#if (OPENSSL_VERSION_NUMBER >= 0x10100000L)
+  if (algorithm == YKPIV_ALGO_ED25519) {
+
+    // Generate a dummy ED25519 to sign with OpenSSL
+    EVP_PKEY *ed_key = NULL;
+    EVP_PKEY_CTX *ed_ctx = EVP_PKEY_CTX_new_id(EVP_PKEY_ED25519, NULL);
+    EVP_PKEY_keygen_init(ed_ctx);
+    EVP_PKEY_keygen(ed_ctx, &ed_key);
+    EVP_PKEY_CTX_free(ed_ctx);
+
+    // Sign the X509 object using the dummy key
+    if (X509_sign(x509, ed_key, md) == 0) {
+      fprintf(stderr, "Failed signing certificate.\n");
+      ERR_print_errors_fp(stderr);
+      EVP_PKEY_free(ed_key);
+      goto selfsign_out;
+    }
+    EVP_PKEY_free(ed_key);
+
+    // Extract the certificate data without the signature
+    unsigned char *tbs_data = NULL;
+    int tbs_len = i2d_re_X509_tbs(x509, &tbs_data);
+
+    // Sign the certificate data using the YubiKey
+    unsigned char yk_sig[64] = {0};
+    size_t yk_siglen = sizeof(yk_sig);
+    if (!sign_data(state, tbs_data, tbs_len, yk_sig, &yk_siglen, algorithm, key)) {
+      fprintf(stderr, "Failed signing tbs certificate portion.\n");
+      goto selfsign_out;
+    }
+
+    // Replace the dummy signature with the signature from the yubikey
+    ASN1_BIT_STRING *psig;
+    const X509_ALGOR *palg;
+    X509_get0_signature((const ASN1_BIT_STRING **) &psig, &palg, x509);
+    ASN1_BIT_STRING_set(psig, yk_sig, yk_siglen);
+  } else {
+#endif
+    /* With opaque structures we can not touch whatever we want, but we need
+     * to embed the sign_data function in the RSA/EC key structures  */
+    EVP_PKEY *sk = wrap_public_key(state, algorithm, public_key, key, oid, oid_len);
+
+    if(X509_sign(x509, sk, md) == 0) {
+      fprintf(stderr, "Failed signing certificate.\n");
+      ERR_print_errors_fp(stderr);
+      EVP_PKEY_free(sk);
+      goto selfsign_out;
+    }
     EVP_PKEY_free(sk);
-    goto selfsign_out;
+#if (OPENSSL_VERSION_NUMBER >= 0x10100000L)
   }
+#endif
 
-  EVP_PKEY_free(sk);
 #endif
 
   if(key_format == key_format_arg_PEM) {
@@ -1227,9 +1449,14 @@ static bool selfsign_certificate(ykpiv_state *state, enum enum_key_format key_fo
     } else {
       fprintf(stderr, "Failed writing x509 information\n");
     }
-
+  } else if(key_format == key_format_arg_DER) {
+    if(i2d_X509_fp(output_file, x509)) {
+      ret = true;
+    } else {
+      fprintf(stderr, "Failed writing DER information\n");
+    }
   } else {
-    fprintf(stderr, "Only PEM support available for certificates.\n");
+    fprintf(stderr, "Only PEM and DER support available for certificates.\n");
   }
 
 selfsign_out:
@@ -1415,10 +1642,10 @@ static bool sign_file(ykpiv_state *state, const char *input, const char *output,
   FILE *output_file = NULL;
   int key;
   unsigned int hash_len;
-  unsigned char hashed[EVP_MAX_MD_SIZE * 2] = {0};
+  unsigned char hashed[YKPIV_OBJ_MAX_SIZE] = {0};
   bool ret = false;
   int algo;
-  const EVP_MD *md;
+  const EVP_MD *md = NULL;
 
   key = get_slot_hex(slot);
 
@@ -1446,35 +1673,45 @@ static bool sign_file(ykpiv_state *state, const char *input, const char *output,
 
   {
     EVP_MD_CTX *mdctx;
-
-    md = get_hash(hash, NULL, NULL);
-    if(md == NULL) {
+    if(algo == YKPIV_ALGO_X25519) {
+      fprintf(stderr, "Signing with X25519 key is not supported\n");
       goto out;
-    }
-
-    mdctx = EVP_MD_CTX_create();
-    if(EVP_DigestInit_ex(mdctx, md, NULL) != 1) {
-      fprintf(stderr, "Failed to initialize digest operation\n");
-      goto out;
-    }
-    while(!feof(input_file)) {
-      char buf[1024] = {0};
-      size_t len = fread(buf, 1, 1024, input_file);
-      if(EVP_DigestUpdate(mdctx, buf, len) != 1) {
-        fprintf(stderr, "Failed to update digest data\n");
+    } else if (algo == YKPIV_ALGO_ED25519) {
+      hash_len = fread(hashed, 1, sizeof(hashed), input_file);
+      if(hash_len >= sizeof(hashed)) {
+        fprintf(stderr, "Cannot perform signature. File too big.\n");
+        goto out;
+      }
+    } else {
+      md = get_hash(hash, NULL, NULL);
+      if (md == NULL) {
         goto out;
       }
-    }
-    if(EVP_DigestFinal_ex(mdctx, hashed, &hash_len) != 1) {
-      fprintf(stderr, "Failed to finalize digest operation\n");
-      goto out;
-    }
 
-    if(verbosity) {
-      fprintf(stderr, "File hashed as: ");
-      dump_data(hashed, hash_len, stderr, true, format_arg_hex);
+      mdctx = EVP_MD_CTX_create();
+      if (EVP_DigestInit_ex(mdctx, md, NULL) != 1) {
+        fprintf(stderr, "Failed to initialize digest operation\n");
+        goto out;
+      }
+      while (!feof(input_file)) {
+        char buf[8192] = {0};
+        size_t len = fread(buf, 1, sizeof(buf), input_file);
+        if (EVP_DigestUpdate(mdctx, buf, len) != 1) {
+          fprintf(stderr, "Failed to update digest data\n");
+          goto out;
+        }
+      }
+      if (EVP_DigestFinal_ex(mdctx, hashed, &hash_len) != 1) {
+        fprintf(stderr, "Failed to finalize digest operation\n");
+        goto out;
+      }
+
+      if (verbosity) {
+        fprintf(stderr, "File hashed as: ");
+        dump_data(hashed, hash_len, stderr, true, format_arg_hex);
+      }
+      EVP_MD_CTX_destroy(mdctx);
     }
-    EVP_MD_CTX_destroy(mdctx);
   }
 
   if(YKPIV_IS_RSA(algo)) {
@@ -1514,9 +1751,7 @@ static void print_cert_info(ykpiv_state *state, enum enum_slot slot, const EVP_M
   int object = (int)ykpiv_util_slot_object(get_slot_hex(slot));
   int slot_name;
   unsigned char data[YKPIV_OBJ_MAX_SIZE] = {0};
-  const unsigned char *ptr = data;
   unsigned long len = sizeof(data);
-  unsigned long offs, cert_len;
   X509 *x509 = NULL;
   X509_NAME *subj;
   BIO *bio = NULL;
@@ -1529,98 +1764,107 @@ static void print_cert_info(ykpiv_state *state, enum enum_slot slot, const EVP_M
 
   fprintf(output, "Slot %x:\t", slot_name);
 
-  if(*ptr++ == TAG_CERT) {
-    unsigned int md_len = sizeof(data);
-    const ASN1_TIME *not_before, *not_after;
+  unsigned char certdata[YKPIV_OBJ_MAX_SIZE * 10] = {0};
+  size_t certdata_len = sizeof(certdata);
+  if(ykpiv_util_get_certdata(data, len, certdata, &certdata_len) != YKPIV_OK) {
+    fprintf(output, "Failed to get certificate data\n");
+    return;
+  }
 
-    offs = get_length(ptr, data + len, &cert_len);
-    if(!offs) {
-      fprintf(output, "Invalid cert length.\n");
-      goto cert_out;
-    }
-    ptr += offs;
-    x509 = d2i_X509(NULL, &ptr, cert_len);
-    if(!x509) {
-      fprintf(output, "Invalid cert data.\n");
-      goto cert_out;
-    }
-    {
-      EVP_PKEY *key = X509_get_pubkey(x509);
-      if(!key) {
-        fprintf(output, "Parse error.\n");
-        goto cert_out;
-      }
-      fprintf(output, "\n\tAlgorithm:\t");
-      switch(get_algorithm(key)) {
-        case YKPIV_ALGO_RSA1024:
-          fprintf(output, "RSA1024\n");
-          break;
-        case YKPIV_ALGO_RSA2048:
-          fprintf(output, "RSA2048\n");
-          break;
-        case YKPIV_ALGO_ECCP256:
-          fprintf(output, "ECCP256\n");
-          break;
-        case YKPIV_ALGO_ECCP384:
-          fprintf(output, "ECCP384\n");
-          break;
-        default:
-          fprintf(output, "Unknown\n");
-      }
-      EVP_PKEY_free(key);
-    }
-    subj = X509_get_subject_name(x509);
-    if(!subj) {
-      fprintf(output, "Parse error.\n");
-      goto cert_out;
-    }
-    fprintf(output, "\tSubject DN:\t");
-    if(X509_NAME_print_ex_fp(output, subj, 0, XN_FLAG_COMPAT) != 1) {
-      fprintf(output, "Failed to write Subject DN.\n");
-      goto cert_out;
-    }
-    fprintf(output, "\n");
-    subj = X509_get_issuer_name(x509);
-    if(!subj) {
-      fprintf(output, "Parse error.\n");
-      goto cert_out;
-    }
-    fprintf(output, "\tIssuer DN:\t");
-    if(X509_NAME_print_ex_fp(output, subj, 0, XN_FLAG_COMPAT) != 1) {
-      fprintf(output, "Failed to write Issuer DN.\n");
-      goto cert_out;
-    }
-    fprintf(output, "\n");
-    if(X509_digest(x509, md, data, &md_len) != 1) {
-      fprintf(output, "Failed to digest data.\n");
-      goto cert_out;
-    }
-    fprintf(output, "\tFingerprint:\t");
-    dump_data(data, md_len, output, false, format_arg_hex);
-
-    bio = BIO_new_fp(output, BIO_NOCLOSE | BIO_FP_TEXT);
-    not_before = X509_get_notBefore(x509);
-    if(not_before) {
-      fprintf(output, "\tNot Before:\t");
-      if(ASN1_TIME_print(bio, not_before) != 1) {
-        fprintf(output, "Failed to write Not Before time.\n");
-        goto cert_out;
-      }
-      fprintf(output, "\n");
-    }
-    not_after = X509_get_notAfter(x509);
-    if(not_after) {
-      fprintf(output, "\tNot After:\t");
-      if(ASN1_TIME_print(bio, not_after) != 1) {
-        fprintf(output, "Failed to write Not After time.\n");
-        goto cert_out;
-      }
-      fprintf(output, "\n");
-    }
-  } else {
+  const unsigned char *certdata_ptr = certdata;
+  x509 = d2i_X509(NULL, &certdata_ptr, certdata_len);
+  if (x509 == NULL) {
     fprintf(output, "Parse error.\n");
     return;
   }
+
+  unsigned int md_len = sizeof(data);
+  const ASN1_TIME *not_before, *not_after;
+
+  EVP_PKEY *key = X509_get_pubkey(x509);
+  if(!key) {
+    fprintf(output, "Parse error.\n");
+    goto cert_out;
+  }
+  fprintf(output, "\n\tAlgorithm:\t");
+  switch(get_algorithm(key)) {
+    case YKPIV_ALGO_RSA1024:
+      fprintf(output, "RSA1024\n");
+      break;
+    case YKPIV_ALGO_RSA2048:
+      fprintf(output, "RSA2048\n");
+      break;
+    case YKPIV_ALGO_RSA3072:
+      fprintf(output, "RSA3072\n");
+      break;
+    case YKPIV_ALGO_RSA4096:
+      fprintf(output, "RSA4096\n");
+      break;
+    case YKPIV_ALGO_ECCP256:
+      fprintf(output, "ECCP256\n");
+      break;
+    case YKPIV_ALGO_ECCP384:
+      fprintf(output, "ECCP384\n");
+      break;
+    case YKPIV_ALGO_ED25519:
+      fprintf(output, "ED25519\n");
+      break;
+    case YKPIV_ALGO_X25519:
+      fprintf(output, "X25519\n");
+      break;
+    default:
+      fprintf(output, "Unknown\n");
+  }
+  EVP_PKEY_free(key);
+
+  subj = X509_get_subject_name(x509);
+  if(!subj) {
+    fprintf(output, "Parse error.\n");
+    goto cert_out;
+  }
+  fprintf(output, "\tSubject DN:\t");
+  if(X509_NAME_print_ex_fp(output, subj, 0, XN_FLAG_COMPAT) != 1) {
+    fprintf(output, "Failed to write Subject DN.\n");
+    goto cert_out;
+  }
+  fprintf(output, "\n");
+  subj = X509_get_issuer_name(x509);
+  if(!subj) {
+    fprintf(output, "Parse error.\n");
+    goto cert_out;
+  }
+  fprintf(output, "\tIssuer DN:\t");
+  if(X509_NAME_print_ex_fp(output, subj, 0, XN_FLAG_COMPAT) != 1) {
+    fprintf(output, "Failed to write Issuer DN.\n");
+    goto cert_out;
+  }
+  fprintf(output, "\n");
+  if(X509_digest(x509, md, data, &md_len) != 1) {
+    fprintf(output, "Failed to digest data.\n");
+    goto cert_out;
+  }
+  fprintf(output, "\tFingerprint:\t");
+  dump_data(data, md_len, output, false, format_arg_hex);
+
+  bio = BIO_new_fp(output, BIO_NOCLOSE | BIO_FP_TEXT);
+  not_before = X509_get_notBefore(x509);
+  if(not_before) {
+    fprintf(output, "\tNot Before:\t");
+    if(ASN1_TIME_print(bio, not_before) != 1) {
+      fprintf(output, "Failed to write Not Before time.\n");
+      goto cert_out;
+    }
+    fprintf(output, "\n");
+  }
+  not_after = X509_get_notAfter(x509);
+  if(not_after) {
+    fprintf(output, "\tNot After:\t");
+    if(ASN1_TIME_print(bio, not_after) != 1) {
+      fprintf(output, "Failed to write Not After time.\n");
+      goto cert_out;
+    }
+    fprintf(output, "\n");
+  }
 cert_out:
   if(x509) {
     X509_free(x509);
@@ -1796,6 +2040,8 @@ static bool test_signature(ykpiv_state *state, enum enum_slot slot,
     switch(algorithm) {
       case YKPIV_ALGO_RSA1024:
       case YKPIV_ALGO_RSA2048:
+      case YKPIV_ALGO_RSA3072:
+      case YKPIV_ALGO_RSA4096:
         {
           RSA *rsa = EVP_PKEY_get1_RSA(pubkey);
           if(!rsa) {
@@ -2173,6 +2419,7 @@ int main(int argc, char *argv[]) {
       case action_arg_testMINUS_signature:
       case action_arg_testMINUS_decipher:
       case action_arg_attest:
+      case action_arg_deleteMINUS_key:
         if(args_info.slot_arg == slot__NULL) {
           fprintf(stderr, "The '%s' action needs a slot (-s) to operate on.\n",
               cmdline_parser_action_values[action]);
@@ -2180,6 +2427,14 @@ int main(int argc, char *argv[]) {
           return EXIT_FAILURE;
         }
         break;
+      case action_arg_moveMINUS_key:
+        if(args_info.slot_arg == slot__NULL || args_info.to_slot_arg == to_slot__NULL) {
+          fprintf(stderr, "The '%s' action needs both a slot (-s) to operate on and a --to-slot to move the key to.\n",
+                  cmdline_parser_action_values[action]);
+          cmdline_parser_free(&args_info);
+          return EXIT_FAILURE;
+        }
+        break;
       case action_arg_pinMINUS_retries:
         if(!args_info.pin_retries_given || !args_info.puk_retries_given) {
           fprintf(stderr, "The '%s' action needs both --pin-retries and --puk-retries arguments.\n",
@@ -2268,6 +2523,8 @@ int main(int argc, char *argv[]) {
       case action_arg_setMINUS_ccc:
       case action_arg_deleteMINUS_certificate:
       case action_arg_writeMINUS_object:
+      case action_arg_moveMINUS_key:
+      case action_arg_deleteMINUS_key:
         if(!authed) {
           if(verbosity) {
             fprintf(stderr, "Authenticating since action '%s' needs that.\n", cmdline_parser_action_values[action]);
@@ -2416,7 +2673,7 @@ int main(int argc, char *argv[]) {
         }
         break;
       case action_arg_importMINUS_certificate:
-        if(import_cert(state, args_info.key_format_arg, args_info.input_arg, args_info.slot_arg, password) == false) {
+        if(import_cert(state, args_info.key_format_arg, args_info.compress_flag, args_info.input_arg, args_info.slot_arg, password) == false) {
           ret = EXIT_FAILURE;
         } else {
           fprintf(stderr, "Successfully imported a new certificate.\n");
@@ -2560,6 +2817,23 @@ int main(int argc, char *argv[]) {
           ret = EXIT_FAILURE;
         }
         break;
+      case action_arg_moveMINUS_key: {
+        int from_slot = get_slot_hex(args_info.slot_arg);
+        int to_slot = get_slot_hex((enum enum_slot) args_info.to_slot_arg);
+        if (move_key(state, from_slot, to_slot) == false) {
+          ret = EXIT_FAILURE;
+        } else {
+          fprintf(stderr, "Successfully moved key.\n");
+        }
+        break;
+      }
+      case action_arg_deleteMINUS_key:
+        if(move_key(state, get_slot_hex(args_info.slot_arg), 0xFF) == false) {
+          ret = EXIT_FAILURE;
+        } else {
+          fprintf(stderr, "Successfully deleted key.\n");
+        }
+        break;
       case action__NULL:
       default:
         fprintf(stderr, "Wrong action. %d.\n", action);
