@@ -107,26 +107,20 @@ rsa_enc_cleanup:
 
 CK_RV do_store_cert(CK_BYTE_PTR data, CK_ULONG len, ykcs11_x509_t **cert) {
 
-  const unsigned char *p = data; // Mandatory temp variable required by OpenSSL
-  unsigned long       offs, cert_len;
+  unsigned char certdata[YKPIV_OBJ_MAX_SIZE * 10] = {0};
+  size_t certdata_len = sizeof (certdata);
 
-  if (*p == TAG_CERT) {
-    // The certificate is in "PIV" format 0x70 len 0x30 len ...
-    p++;
-    offs = get_length(p, data + len, &cert_len);
-    if(!offs)
-      return CKR_ARGUMENTS_BAD;
-    p += offs;
-  }
-  else {
-    // Raw certificate ...
-    cert_len = len;
+  if(ykpiv_util_get_certdata(data, len, certdata, &certdata_len) != YKPIV_OK) {
+    DBG("Failed to get certificate data");
+    return CKR_DATA_INVALID;
   }
 
-  if(*cert)
+  if(*cert) {
     X509_free(*cert);
+  }
 
-  *cert = d2i_X509(NULL, &p, cert_len);
+  const unsigned char *p = certdata; // Mandatory temp variable required by OpenSSL
+  *cert = d2i_X509(NULL, &p, certdata_len);
   if (*cert == NULL)
     return CKR_FUNCTION_FAILED;
 
@@ -565,6 +559,10 @@ CK_BYTE do_get_key_algorithm(ykcs11_pkey_t *key) {
         return YKPIV_ALGO_RSA1024;
       case 2048:
         return YKPIV_ALGO_RSA2048;
+      case 3072:
+        return YKPIV_ALGO_RSA3072;
+      case 4096:
+        return YKPIV_ALGO_RSA4096;
       }
     case EVP_PKEY_EC:
       switch(EVP_PKEY_bits(key)) {
@@ -594,20 +592,33 @@ CK_RV do_get_modulus(ykcs11_pkey_t *key, CK_BYTE_PTR data, CK_ULONG len) {
   return CKR_OK;
 }
 
-CK_RV do_get_public_exponent(ykcs11_pkey_t *key, CK_BYTE_PTR data, CK_ULONG len) {
+CK_BBOOL do_check_public_exponent(CK_BYTE_PTR data, CK_ULONG len) {
+  BIGNUM *bn = BN_bin2bn(data, len, NULL);
+  BIGNUM *f4 = BN_new();
+  BN_set_word(f4, 0x10001);
+  CK_BBOOL ret = BN_cmp(bn, f4) ? CK_FALSE : CK_TRUE;
+  BN_free(f4);
+  BN_free(bn);
+  return ret;
+}
 
-  const RSA *rsa = NULL;
-  const BIGNUM *bn_e;
+CK_RV do_get_public_exponent(ykcs11_pkey_t *key, CK_BYTE_PTR data, CK_ULONG_PTR len) {
 
-  rsa = key ? EVP_PKEY_get0_RSA(key) : 0;
+  const RSA *rsa = key ? EVP_PKEY_get0_RSA(key) : 0;
   if (rsa == NULL)
     return CKR_ATTRIBUTE_TYPE_INVALID;
 
+  const BIGNUM *bn_e = NULL;
   RSA_get0_key(rsa, NULL, &bn_e, NULL);
 
-  if(BN_bn2binpad(bn_e, data, len) < 0)
+  if (bn_e == NULL)
+    return CKR_ATTRIBUTE_TYPE_INVALID;
+
+  if(*len < BN_num_bytes(bn_e))
     return CKR_DATA_LEN_RANGE;
 
+  *len = BN_bn2bin(bn_e, data);
+
   return CKR_OK;
 }
 
