@@ -60,7 +60,8 @@ CK_ULONG    do_get_key_bits(ykcs11_pkey_t *key);
 CK_ULONG    do_get_key_size(ykcs11_pkey_t *key);
 CK_ULONG    do_get_signature_size(ykcs11_pkey_t *key);
 CK_BYTE     do_get_key_algorithm(ykcs11_pkey_t *key);
-CK_RV       do_get_public_exponent(ykcs11_pkey_t *key, CK_BYTE_PTR data, CK_ULONG len);
+CK_BBOOL    do_check_public_exponent(CK_BYTE_PTR data, CK_ULONG len);
+CK_RV       do_get_public_exponent(ykcs11_pkey_t *key, CK_BYTE_PTR data, CK_ULONG_PTR len);
 CK_RV       do_get_public_key(ykcs11_pkey_t *key, CK_BYTE_PTR data, CK_ULONG_PTR len);
 CK_RV       do_get_modulus(ykcs11_pkey_t *key, CK_BYTE_PTR data, CK_ULONG len);
 CK_RV       do_get_curve_parameters(ykcs11_pkey_t *key, CK_BYTE_PTR data, CK_ULONG_PTR len);
