@@ -39,7 +39,7 @@
 #include "../common/util.h"
 
 #define MIN_RSA_KEY_SIZE 1024
-#define MAX_RSA_KEY_SIZE 2048
+#define MAX_RSA_KEY_SIZE 4096
 #define MIN_ECC_KEY_SIZE 256
 #define MAX_ECC_KEY_SIZE 384
 
@@ -59,14 +59,15 @@ static const token_mechanism token_mechanisms[] = {
   CKM_SHA256_RSA_PKCS_PSS, {MIN_RSA_KEY_SIZE, MAX_RSA_KEY_SIZE, CKF_HW | CKF_SIGN | CKF_VERIFY},
   CKM_SHA384_RSA_PKCS_PSS, {MIN_RSA_KEY_SIZE, MAX_RSA_KEY_SIZE, CKF_HW | CKF_SIGN | CKF_VERIFY},
   CKM_SHA512_RSA_PKCS_PSS, {MIN_RSA_KEY_SIZE, MAX_RSA_KEY_SIZE, CKF_HW | CKF_SIGN | CKF_VERIFY},
-  CKM_EC_KEY_PAIR_GEN, {MIN_ECC_KEY_SIZE, MAX_ECC_KEY_SIZE, CKF_HW | CKF_GENERATE_KEY_PAIR},
-  //CKM_ECDSA_KEY_PAIR_GEN, {MIN_ECC_KEY_SIZE, MAX_ECC_KEY_SIZE, CKF_HW | CKF_GENERATE_KEY_PAIR}, //Same as CKM_EC_KEY_PAIR_GEN, deprecated in 2.11
-  CKM_ECDSA, {MIN_ECC_KEY_SIZE, MAX_ECC_KEY_SIZE, CKF_HW | CKF_SIGN | CKF_VERIFY},
-  CKM_ECDSA_SHA1, {MIN_ECC_KEY_SIZE, MAX_ECC_KEY_SIZE, CKF_HW | CKF_SIGN | CKF_VERIFY},
-  CKM_ECDSA_SHA224, {MIN_ECC_KEY_SIZE, MAX_ECC_KEY_SIZE, CKF_HW | CKF_SIGN | CKF_VERIFY},
-  CKM_ECDSA_SHA256, {MIN_ECC_KEY_SIZE, MAX_ECC_KEY_SIZE, CKF_HW | CKF_SIGN | CKF_VERIFY},
-  CKM_ECDSA_SHA384, {MIN_ECC_KEY_SIZE, MAX_ECC_KEY_SIZE, CKF_HW | CKF_SIGN | CKF_VERIFY},
-  CKM_ECDH1_DERIVE, {MIN_ECC_KEY_SIZE, MAX_ECC_KEY_SIZE, CKF_HW | CKF_DERIVE},
+  CKM_EC_KEY_PAIR_GEN, {MIN_ECC_KEY_SIZE, MAX_ECC_KEY_SIZE, CKF_HW | CKF_GENERATE_KEY_PAIR | CKF_EC_F_P | CKF_EC_NAMEDCURVE | CKF_EC_UNCOMPRESS},
+  //CKM_ECDSA_KEY_PAIR_GEN, {MIN_ECC_KEY_SIZE, MAX_ECC_KEY_SIZE, CKF_HW | CKF_GENERATE_KEY_PAIR | CKF_EC_F_P | CKF_EC_NAMEDCURVE | CKF_EC_UNCOMPRESS}, //Same as CKM_EC_KEY_PAIR_GEN, deprecated in 2.11
+  CKM_ECDSA, {MIN_ECC_KEY_SIZE, MAX_ECC_KEY_SIZE, CKF_HW | CKF_SIGN | CKF_VERIFY | CKF_EC_F_P | CKF_EC_NAMEDCURVE | CKF_EC_UNCOMPRESS},
+  CKM_ECDSA_SHA1, {MIN_ECC_KEY_SIZE, MAX_ECC_KEY_SIZE, CKF_HW | CKF_SIGN | CKF_VERIFY | CKF_EC_F_P | CKF_EC_NAMEDCURVE | CKF_EC_UNCOMPRESS},
+  CKM_ECDSA_SHA224, {MIN_ECC_KEY_SIZE, MAX_ECC_KEY_SIZE, CKF_HW | CKF_SIGN | CKF_VERIFY | CKF_EC_F_P | CKF_EC_NAMEDCURVE | CKF_EC_UNCOMPRESS},
+  CKM_ECDSA_SHA256, {MIN_ECC_KEY_SIZE, MAX_ECC_KEY_SIZE, CKF_HW | CKF_SIGN | CKF_VERIFY | CKF_EC_F_P | CKF_EC_NAMEDCURVE | CKF_EC_UNCOMPRESS},
+  CKM_ECDSA_SHA384, {MIN_ECC_KEY_SIZE, MAX_ECC_KEY_SIZE, CKF_HW | CKF_SIGN | CKF_VERIFY | CKF_EC_F_P | CKF_EC_NAMEDCURVE | CKF_EC_UNCOMPRESS},
+  CKM_ECDSA_SHA512, {MIN_ECC_KEY_SIZE, MAX_ECC_KEY_SIZE, CKF_HW | CKF_SIGN | CKF_VERIFY | CKF_EC_F_P | CKF_EC_NAMEDCURVE | CKF_EC_UNCOMPRESS},
+  CKM_ECDH1_DERIVE, {MIN_ECC_KEY_SIZE, MAX_ECC_KEY_SIZE, CKF_HW | CKF_DERIVE | CKF_EC_F_P | CKF_EC_NAMEDCURVE | CKF_EC_UNCOMPRESS},
   CKM_SHA_1, {0, 0, CKF_DIGEST},
   CKM_SHA256, {0, 0, CKF_DIGEST},
   CKM_SHA384, {0, 0, CKF_DIGEST},
@@ -141,14 +142,15 @@ CK_RV get_token_model(ykpiv_state *state, CK_UTF8CHAR_PTR str, CK_ULONG len) {
 CK_RV get_token_version(ykpiv_state *state, CK_VERSION_PTR version) {
 
   char buf[16] = {0};
+  ykpiv_rc rc;
 
   if (version == NULL)
     return CKR_ARGUMENTS_BAD;
 
-  if (ykpiv_get_version(state, buf, sizeof(buf)) != YKPIV_OK) {
+  if ((rc = ykpiv_get_version(state, buf, sizeof(buf))) != YKPIV_OK) {
     version->major = 0;
     version->minor = 0;
-    return CKR_DEVICE_ERROR;
+    return yrc_to_rv(rc);
   }
 
   version->major = (buf[0] - '0');
@@ -159,44 +161,42 @@ CK_RV get_token_version(ykpiv_state *state, CK_VERSION_PTR version) {
 
 CK_RV get_token_serial(ykpiv_state *state, CK_CHAR_PTR str, CK_ULONG len) {
 
-  uint32_t serial;
+  uint32_t serial = 0;
   char buf[64] = {0};
-  int actual;
 
   ykpiv_rc rc = ykpiv_get_serial(state, &serial);
 
-  actual = snprintf(buf, sizeof(buf), "%u", serial);
+  int actual = snprintf(buf, sizeof(buf), "%u", serial);
 
   if(actual < 0)
     return CKR_FUNCTION_FAILED;
 
-  if(actual >= len)
+  if((CK_ULONG)actual >= len)
     return CKR_BUFFER_TOO_SMALL;
 
   memstrcpy(str, len, buf);
 
-  return rc == YKPIV_OK ? CKR_OK : CKR_DEVICE_ERROR;
+  return yrc_to_rv(rc);
 }
 
 CK_RV get_token_label(ykpiv_state *state, CK_CHAR_PTR str, CK_ULONG len) {
 
-  uint32_t serial;
+  uint32_t serial = 0;
   char buf[64] = {0};
-  int actual;
 
   ykpiv_rc rc = ykpiv_get_serial(state, &serial);
 
-  actual = snprintf(buf, sizeof(buf), "YubiKey PIV #%u", serial);
+  int actual = snprintf(buf, sizeof(buf), "YubiKey PIV #%u", serial);
 
   if(actual < 0)
     return CKR_FUNCTION_FAILED;
 
-  if(actual >= len)
+  if((CK_ULONG)actual >= len)
     return CKR_BUFFER_TOO_SMALL;
 
   memstrcpy(str, len, buf);
 
-  return rc == YKPIV_OK ? CKR_OK : CKR_DEVICE_ERROR;
+  return yrc_to_rv(rc);
 }
 
 CK_RV get_token_mechanism_list(CK_MECHANISM_TYPE_PTR mec, CK_ULONG_PTR num) {
@@ -241,24 +241,37 @@ CK_RV token_change_pin(ykpiv_state *state, CK_USER_TYPE user_type, CK_UTF8CHAR_P
 
   switch(user_type){
     case CKU_SO:{
-      unsigned char new_key[24] = {0};
-      size_t new_key_len = sizeof(new_key);
-      if(ykpiv_hex_decode((const char*)pNewPin, ulNewLen, new_key, &new_key_len) != YKPIV_OK) {
-        DBG("Failed to decode new pin")
+      ykpiv_mgm new_key = {0};
+      new_key.len = sizeof(new_key.data);
+      if(ykpiv_hex_decode((const char*)pNewPin, ulNewLen, new_key.data, &new_key.len) != YKPIV_OK) {
+        DBG("Failed to decode new pin");
         return CKR_PIN_INVALID;
       }
-      DBG("Changing SO PIN")
-      res = ykpiv_set_mgmkey(state, new_key);
-      OPENSSL_cleanse(new_key, sizeof(new_key));
+      DBG("Changing SO PIN");
+      // Set new mgm key (SO PIN) with the same algorithm and touch policy the old one had
+      res = ykpiv_set_mgmkey3(state, new_key.data, new_key.len, YKPIV_ALGO_AUTO, YKPIV_TOUCHPOLICY_AUTO);
+      if(res == YKPIV_OK) {
+        ykpiv_config config = {0};
+        res = ykpiv_util_get_config(state, &config);
+        if(res == YKPIV_OK && config.mgm_type == YKPIV_CONFIG_MGM_PROTECTED) {
+          res = ykpiv_util_update_protected_mgm(state, &new_key);
+          if(res != YKPIV_OK) {
+            DBG("Failed to update pin protected management key metadata");
+          }
+        }
+      } else {
+        DBG("Failed to set new management key: %s", ykpiv_strerror(res));
+      }
+      OPENSSL_cleanse(new_key.data, sizeof(new_key.data));
       break;
     }
     case CKU_USER:
       if(ulOldLen >= 4 && strncmp((const char*)pOldPin, "puk:", 4) == 0){
         if(ulNewLen >= 4 && strncmp((const char*)pNewPin, "pin:", 4) == 0) {
-          DBG("Unblocking PIN with PUK")
+          DBG("Unblocking PIN with PUK");
           res = ykpiv_unblock_pin(state, (const char*)pOldPin + 4, ulOldLen - 4, (const char*)pNewPin + 4, ulNewLen - 4, &tries);
         } else {
-          DBG("Changing PUK")
+          DBG("Changing PUK");
           if(ulNewLen >= 4 && strncmp((const char*)pNewPin, "puk:", 4) == 0) {
             res = ykpiv_change_puk(state, (const char*)pOldPin + 4, ulOldLen - 4, (const char*)pNewPin + 4, ulNewLen - 4, &tries);
           } else {
@@ -266,7 +279,7 @@ CK_RV token_change_pin(ykpiv_state *state, CK_USER_TYPE user_type, CK_UTF8CHAR_P
           }
         }
       }else{
-        DBG("Changing PIN")
+        DBG("Changing PIN");
         res = ykpiv_change_pin(state, (const char*)pOldPin, ulOldLen, (const char*)pNewPin, ulNewLen, &tries);
       }
       break;
@@ -276,16 +289,10 @@ CK_RV token_change_pin(ykpiv_state *state, CK_USER_TYPE user_type, CK_UTF8CHAR_P
   }
 
   switch (res) {
-    case YKPIV_OK:
-      return CKR_OK;
     case YKPIV_SIZE_ERROR:
       return CKR_PIN_LEN_RANGE;
-    case YKPIV_WRONG_PIN:
-      return CKR_PIN_INCORRECT;
-    case YKPIV_PIN_LOCKED:
-      return CKR_PIN_LOCKED;
     default:
-      return CKR_DEVICE_ERROR;
+      return yrc_to_rv(res);
   }
 }
 
@@ -299,7 +306,7 @@ CK_RV token_login(ykpiv_state *state, CK_USER_TYPE user, CK_UTF8CHAR_PTR pin, CK
     memcpy(term_pin, pin, pin_len);
     term_pin[pin_len] = 0;
 
-    int tries;
+    int tries = 0;
     res = ykpiv_verify(state, term_pin, &tries);
 
     OPENSSL_cleanse(term_pin, pin_len);
@@ -307,88 +314,81 @@ CK_RV token_login(ykpiv_state *state, CK_USER_TYPE user, CK_UTF8CHAR_PTR pin, CK
     if (res != YKPIV_OK) {
       DBG("Failed to login: %s, %d tries left", ykpiv_strerror(res), tries);
 
-      if(res == YKPIV_WRONG_PIN)
-        return CKR_PIN_INCORRECT;
-
-      if(res == YKPIV_PIN_LOCKED)
-        return CKR_PIN_LOCKED;
-
-      return CKR_DEVICE_ERROR;
+      return yrc_to_rv(res);
     }
-  } else if(pin_len != YKPIV_MGM_KEY_LEN || user != CKU_SO) {
+  } else if(pin_len < YKPIV_MIN_MGM_KEY_LEN || pin_len > YKPIV_MAX_MGM_KEY_LEN || user != CKU_SO) {
     DBG("PIN is wrong length");
     return CKR_ARGUMENTS_BAD;
   }
 
   if (user == CKU_SO) {
-    unsigned char key[24] = {0};
+    ykpiv_config cfg = {0};
 
-    if (pin_len == YKPIV_MGM_KEY_LEN) {
-      size_t key_len = sizeof(key);
-      if(ykpiv_hex_decode((char *)pin, pin_len, key, &key_len) != YKPIV_OK) {
+    if (pin_len >= YKPIV_MIN_MGM_KEY_LEN && pin_len <= YKPIV_MAX_MGM_KEY_LEN) {
+      cfg.mgm_len = sizeof(cfg.mgm_key);
+      if((res = ykpiv_hex_decode((char *)pin, pin_len, cfg.mgm_key, &cfg.mgm_len)) != YKPIV_OK) {
         DBG("Failed decoding key");
-        OPENSSL_cleanse(key, key_len);
-        return CKR_ARGUMENTS_BAD;
+        OPENSSL_cleanse(cfg.mgm_key, sizeof(cfg.mgm_key));
+        return yrc_to_rv(res);
       }
     } else {
-      ykpiv_config cfg;
       res = ykpiv_util_get_config(state, &cfg);
       if(res != YKPIV_OK) {
         DBG("Failed to get device configuration: %s", ykpiv_strerror(res));
-        return CKR_DEVICE_ERROR;
+        OPENSSL_cleanse(cfg.mgm_key, sizeof(cfg.mgm_key));
+        return yrc_to_rv(res);
       }
 
       if(cfg.mgm_type != YKPIV_CONFIG_MGM_PROTECTED) {
         DBG("Device configuration invalid, no PIN-protected MGM key available");
+        OPENSSL_cleanse(cfg.mgm_key, sizeof(cfg.mgm_key));
         return CKR_USER_PIN_NOT_INITIALIZED;
       }
-
-      memcpy(key, cfg.mgm_key, sizeof(key));
-      OPENSSL_cleanse(cfg.mgm_key, sizeof(cfg.mgm_key));
     }
 
-    if((res = ykpiv_authenticate(state, key)) != YKPIV_OK) {
+    if((res = ykpiv_authenticate2(state, cfg.mgm_key, cfg.mgm_len)) != YKPIV_OK) {
       DBG("Failed to authenticate: %s", ykpiv_strerror(res));
-      OPENSSL_cleanse(key, sizeof(key));
+      OPENSSL_cleanse(cfg.mgm_key, sizeof(cfg.mgm_key));
 
       if(res == YKPIV_AUTHENTICATION_ERROR)
         return CKR_PIN_INCORRECT;
 
-      return CKR_DEVICE_ERROR;
+      return yrc_to_rv(res);
     }
 
-    OPENSSL_cleanse(key, sizeof(key));
+    OPENSSL_cleanse(cfg.mgm_key, sizeof(cfg.mgm_key));
   }
 
   return CKR_OK;
 }
 
-CK_RV token_generate_key(ykpiv_state *state, CK_BYTE algorithm, CK_BYTE key, CK_BYTE_PTR cert_data, CK_ULONG_PTR cert_len) {
+CK_RV token_generate_key(ykpiv_state *state, gen_info_t *gen, CK_BYTE key, CK_BYTE_PTR cert_data, CK_ULONG_PTR cert_len) {
   // TODO: make a function in ykpiv for this
   unsigned char in_data[11] = {0};
   unsigned char *in_ptr = in_data;
   unsigned char data[1024] = {0};
   unsigned char templ[] = {0, YKPIV_INS_GENERATE_ASYMMETRIC, 0, 0};
-  unsigned char *certptr;
-  unsigned long len, len_bytes, offs, recv_len = sizeof(data);
-  char version[7] = {0};
+  uint8_t certdata[YKPIV_OBJ_MAX_SIZE + 16] = {0};
+  size_t certdata_len = sizeof(certdata);
+  unsigned long len, offs, recv_len = sizeof(data);
   char label[32] = {0};
+  ykpiv_rc res;
   int sw;
 
-  switch(algorithm) {
+  switch(gen->algorithm) {
     case YKPIV_ALGO_RSA1024:
     case YKPIV_ALGO_RSA2048:
-      if(ykpiv_get_version(state, version, sizeof(version)) == YKPIV_OK) {
-        int major, minor, build;
-        int match = sscanf(version, "%d.%d.%d", &major, &minor, &build);
-        if(match == 3 && major == 4 && (minor < 3 || (minor == 3 && build < 5))) {
-          DBG("On-chip RSA key generation on this YubiKey has been blocked.");
-          DBG("Please see https://yubi.co/ysa201701/ for details.");
-          return CKR_FUNCTION_FAILED;
-        }
-      } else {
-        DBG("Failed to communicate.");
-        return CKR_DEVICE_ERROR;
+    case YKPIV_ALGO_RSA3072:
+    case YKPIV_ALGO_RSA4096:
+      if (!is_version_compatible(state, 4, 3, 5)) {
+        DBG("On-chip RSA key generation on this YubiKey has been blocked.");
+        DBG("Please see https://yubi.co/ysa201701/ for details.");
+        return CKR_FUNCTION_FAILED;
+      }
+      if (!is_version_compatible(state, 5, 7, 0) &&
+         (gen->algorithm == YKPIV_ALGO_RSA3072 || gen->algorithm == YKPIV_ALGO_RSA4096)) {
+        DBG("RSA3072 and RSA4096 key types are only available with YubiKeys with version number 5.7.0 or later");
+        return CKR_FUNCTION_NOT_SUPPORTED;
       }
       break;
 
@@ -400,17 +400,53 @@ CK_RV token_generate_key(ykpiv_state *state, CK_BYTE algorithm, CK_BYTE key, CK_
       return CKR_FUNCTION_FAILED;
   }
 
+  switch (gen->touch_policy) {
+    case YKPIV_TOUCHPOLICY_DEFAULT:
+    case YKPIV_TOUCHPOLICY_ALWAYS:
+    case YKPIV_TOUCHPOLICY_CACHED:
+    case YKPIV_TOUCHPOLICY_NEVER:
+        break;
+    default: 
+        return CKR_FUNCTION_FAILED;
+  }
+
+  switch (gen->pin_policy) {
+    case YKPIV_PINPOLICY_DEFAULT:
+    case YKPIV_PINPOLICY_ALWAYS:
+    case YKPIV_PINPOLICY_ONCE:
+    case YKPIV_PINPOLICY_NEVER:
+        break;
+    default: 
+        return CKR_FUNCTION_FAILED;
+  }
+
   templ[3] = key;
 
   *in_ptr++ = 0xac;
   *in_ptr++ = 3;
   *in_ptr++ = YKPIV_ALGO_TAG;
   *in_ptr++ = 1;
-  *in_ptr++ = algorithm;
+  *in_ptr++ = gen->algorithm;
 
-  if(ykpiv_transfer_data(state, templ, in_data, in_ptr - in_data, data, &recv_len, &sw) != YKPIV_OK || sw != 0x9000) {
-    DBG("Failed to generate key, sw = %04x.", sw);
-    return CKR_DEVICE_ERROR;
+  if (gen->touch_policy != YKPIV_TOUCHPOLICY_DEFAULT) {
+      in_data[1] += 3;
+      *in_ptr++ = YKPIV_TOUCHPOLICY_TAG;
+      *in_ptr++ = 0x01;
+      *in_ptr++ = gen->touch_policy;
+  }
+
+  if (gen->pin_policy != YKPIV_PINPOLICY_DEFAULT) {
+      in_data[1] += 3;
+      *in_ptr++ = YKPIV_PINPOLICY_TAG;
+      *in_ptr++ = 0x01;
+      *in_ptr++ = gen->pin_policy;
+  }
+
+  if((res = ykpiv_transfer_data(state, templ, in_data, in_ptr - in_data, data, &recv_len, &sw)) != YKPIV_OK) {
+    return yrc_to_rv(res);
+  }
+  if((res = ykpiv_translate_sw(sw)) != YKPIV_OK) {
+    return yrc_to_rv(res);
   }
 
   snprintf(label, sizeof(label), "YubiKey PIV Slot %x", key);
@@ -422,35 +458,25 @@ CK_RV token_generate_key(ykpiv_state *state, CK_BYTE algorithm, CK_BYTE key, CK_
 
   len = recv_len;
   recv_len = sizeof(data);
-  CK_RV rv = do_create_empty_cert(data + offs, len - offs, algorithm, label, data, &recv_len);
+  CK_RV rv = do_create_empty_cert(data + offs, len - offs, gen->algorithm, label, data, &recv_len);
   if(rv != CKR_OK)
     return rv;
 
-  len_bytes = get_length_size(recv_len);
+  if ((res = ykpiv_util_write_certdata(data, recv_len, YKPIV_CERTINFO_UNCOMPRESSED, certdata, &certdata_len)) != YKPIV_OK) {
+    return yrc_to_rv(res);
+  }
 
-  certptr = data;
-  memmove(data + len_bytes + 1, data, recv_len);
-
-  *certptr++ = 0x70;
-  certptr += set_length(certptr, recv_len);
-  certptr += recv_len;
-  *certptr++ = 0x71;
-  *certptr++ = 1;
-  *certptr++ = 0; /* certinfo (gzip etc) */
-  *certptr++ = 0xfe; /* LRC */
-  *certptr++ = 0;
-
-  if(*cert_len < certptr - data) {
+  if(*cert_len < (CK_ULONG)certdata_len) {
     DBG("Certificate buffer too small.");
     return CKR_BUFFER_TOO_SMALL;
   }
 
   // Store the certificate into the token
-  if (ykpiv_save_object(state, ykpiv_util_slot_object(key), data, certptr - data) != YKPIV_OK)
-    return CKR_DEVICE_ERROR;
+  if ((res = ykpiv_save_object(state, ykpiv_util_slot_object(key), certdata, certdata_len)) != YKPIV_OK)
+    return yrc_to_rv(res);
 
-  memcpy(cert_data, data, certptr - data);
-  *cert_len = certptr - data;
+  memcpy(cert_data, data, (unsigned long) certdata_len);
+  *cert_len = certdata_len;
 
   return CKR_OK;
 }
@@ -458,90 +484,33 @@ CK_RV token_generate_key(ykpiv_state *state, CK_BYTE algorithm, CK_BYTE key, CK_
 CK_RV token_import_cert(ykpiv_state *state, CK_ULONG cert_id, CK_BYTE_PTR in, CK_ULONG in_len) {
 
   unsigned char certdata[YKPIV_OBJ_MAX_SIZE + 16] = {0};
-  unsigned char *certptr;
+  size_t certdata_len = sizeof(certdata);
   CK_ULONG cert_len;
-
+  ykpiv_rc res;
   CK_RV rv;
 
   // Check whether or not we have a valid cert
-  if ((rv = do_check_cert(in, in_len, &cert_len)) != CKR_OK)
+  if ((rv = do_check_cert(in, in_len, &cert_len)) != CKR_OK) {
+    DBG("Certificate not valid.");
     return rv;
+  }
 
-  if (cert_len > YKPIV_OBJ_MAX_SIZE)
-    return CKR_FUNCTION_FAILED;
-
-  certptr = certdata;
-
-  *certptr++ = 0x70;
-  certptr += set_length(certptr, cert_len);
-  memcpy(certptr, in, cert_len);
-  certptr += cert_len;
-
-  *certptr++ = 0x71;
-  *certptr++ = 1;
-  *certptr++ = 0; /* certinfo (gzip etc) */
-  *certptr++ = 0xfe; /* LRC */
-  *certptr++ = 0;
+  if ((res = ykpiv_util_write_certdata(in, cert_len, YKPIV_CERTINFO_UNCOMPRESSED, certdata, &certdata_len)) != YKPIV_OK) {
+    return yrc_to_rv(res);
+  }
 
   // Store the certificate into the token
-  if (ykpiv_save_object(state, cert_id, certdata, certptr - certdata) != YKPIV_OK)
-    return CKR_DEVICE_ERROR;
-
-  return CKR_OK;
-}
-
-CK_RV token_import_private_key(ykpiv_state *state, CK_BYTE key_id,
-                                      CK_BYTE_PTR p, CK_ULONG p_len,
-                                      CK_BYTE_PTR q, CK_ULONG q_len,
-                                      CK_BYTE_PTR dp, CK_ULONG dp_len,
-                                      CK_BYTE_PTR dq, CK_ULONG dq_len,
-                                      CK_BYTE_PTR qinv, CK_ULONG qinv_len,
-                                      CK_BYTE_PTR ec_data, CK_ULONG ec_data_len) {
-
-  CK_BYTE  pin_policy;
-  CK_BYTE  touch_policy;
-  CK_BYTE  algo;
-  ykpiv_rc rc;
-
-  if (p == NULL) {
-    if (ec_data_len == 32 || ec_data_len == 31)
-      algo = YKPIV_ALGO_ECCP256;
-    else
-      algo = YKPIV_ALGO_ECCP384;
-  }
-  else if (ec_data == NULL) {
-    if (p_len == 64)
-      algo = YKPIV_ALGO_RSA1024;
-    else
-      algo = YKPIV_ALGO_RSA2048;
-  }
-  else
-    return CKR_FUNCTION_FAILED;
-
-  pin_policy = YKPIV_PINPOLICY_DEFAULT;
-  touch_policy = YKPIV_TOUCHPOLICY_DEFAULT;
-
-  rc = ykpiv_import_private_key(state, key_id, algo,
-                                p, p_len,
-                                q, q_len,
-                                dp, dp_len,
-                                dq, dq_len,
-                                qinv, qinv_len,
-                                ec_data, ec_data_len,
-                                pin_policy, touch_policy);
-
-  if (rc != YKPIV_OK) {
-    DBG("ykpiv_import_private_key failed: %s", ykpiv_strerror(rc));
-    return CKR_DEVICE_ERROR;
-  }
+  if ((res = ykpiv_save_object(state, cert_id, certdata, certdata_len)) != YKPIV_OK)
+    return yrc_to_rv(res);
 
   return CKR_OK;
 }
 
 CK_RV token_delete_cert(ykpiv_state *state, CK_ULONG cert_id) {
+  ykpiv_rc res;
 
-  if (ykpiv_save_object(state, cert_id, NULL, 0) != YKPIV_OK)
-    return CKR_DEVICE_ERROR;
+  if ((res = ykpiv_save_object(state, cert_id, NULL, 0)) != YKPIV_OK)
+    return yrc_to_rv(res);
 
   return CKR_OK;
 }
